import collections
import numpy as np
import math
import matplotlib.pyplot as plt
import scipy.sparse
import scipy.sparse.linalg
from dataclasses import dataclass
from typing import List
from typing import Tuple
from typing import Union
from .laminate import Laminate
from .ply import _Lamina
import copy

coefficient = collections.namedtuple("Coefficient", ["pos", "value"])
constraint_def = collections.namedtuple("ConstraintDef", ["value", "cv_list"])


def _finite_difference_coefficients(diff_order: int, stencil_width, h: float) -> List[coefficient]:
    """
    Computes a set of coefficients for differentiation in the finite difference method.

    In the case that diff_order=0, the results in an identity term.

    The stencil_width argument defines how many nodes on either side of the node of interest are used.
    For example, when stencil_width=1, one node ahead and one node behind are used.

    :param diff_order: The order of the derivative desired. This can be 0 or greater
    :param stencil_width: The number of nodes to include on either side of the node of interest (int)
    :param h: The grid spacing (float)
    :return: a list of `coefficient`s, which are a namedtuple of `pos` and `value`
    """
    stencil = np.arange(-stencil_width, stencil_width + 1)
    n = len(stencil)
    assert n >= diff_order
    mat = np.array([
        stencil ** j
        for j in range(n)
    ])
    d = np.zeros((n, 1))
    d[diff_order, 0] = math.factorial(diff_order)
    c = np.linalg.inv(mat) @ d
    return [coefficient(i, float(v / h ** diff_order)) for i, v in zip(stencil, c)]


def matrix_diagnostics(k_sparse):
    k = np.array(k_sparse.todense())

    print("det={}".format(np.linalg.det(k)))
    print("rank={}".format(np.linalg.matrix_rank(k)))

    for i in range(k.shape[0]):
        for j in range(k.shape[0]):
            if j < i:
                inner_product = np.inner(k[i, :], k[j, :])
                norm_i = np.linalg.norm(k[i, :])
                norm_j = np.linalg.norm(k[j, :])
                if np.abs(inner_product - norm_i * norm_j) < 1e-5:
                    print("Rows {} and {} are dependant".format(i, j))


class Model:
    """
    The base class for a model. This class should not be directly instantiated. Sub-classes may be instantiated.
    """
    def __init__(self, prop: Laminate, do_validate: bool = True):
        self.prop = prop
        if do_validate:
            self._validate()

    def _validate(self) -> None:
        et = list(self.equation_terms)
        if len(et) != self.dof:
            raise AssertionError(
                "equation_terms must be a list (or generator) with the number of elements equal to dof"
            )
        for et_i in et:
            if len(et_i) != self.dof:
                raise AssertionError(
                    "Each element within equation_terms must be an iterable with the number of elements equal to dof"
                )

    @property
    def dof(self) -> int:
        """
        The number of degrees of freedom required at each node for the model.
        :return: an integer
        """
        raise NotImplementedError

    @property
    def equation_terms(self) -> List[List[List[Union[float, None]]]]:
        """
        Defines the differential equation(s) of the model.

        This must be an list (or generator) with the number of elements equal to self.dof. Each of these elements
        defines an equation. The first equation is for moment, the second (if self.dof == 2) is shear.

        Each equation has a number of terms equal to self.dof. The first term is for w (vertical deformation), and
        the second (if self.doff == 2) is for theta.

        Each term within the equation is a list constants. The zero-th constant is a multiplier for the term itself
        (eg. w or theta), the first is a multiplier for the first derivative, the second is a multiplier for the
        second derivative, etc. For example, the term [1, 5, 3] would correspond with the equation
        y + 5 dy/dx -3 d^2y/dx^2

        For terms that are empty, None can be given to avoid adding an element to the final stiffness matrix. This
        may help with memory use and speed, compared with specifying 0.
        :return:
        """
        raise NotImplementedError

    def create_force_vector(self, n_model_equations: int, moment: np.array, transverse_force: np.array,
                            h: float) -> np.array:
        """
        Creates a force vector, based on the force and moment distribution. Extra elements are added to allow for
        constrain equations. The total number of elements is n_model_equations.

        :param n_model_equations: The total number of elements in the force vector
        :param moment: A vector of the moment at each node
        :param transverse_force: A vector of the transverse force at each node
        :param h: the spacing between nodes
        :return: A force vector
        """
        raise NotImplementedError

    def constraint_transverse_force(self, i_node: int, h: float, force_magnitude: float) -> List[constraint_def]:
        raise NotImplementedError

    def constraint_pinned(self, i_node: int, h: float) -> List[constraint_def]:
        """
        Creates the equation that defines a pinned constraint.

        :param i_node: the index of the node to constrain
        :param h: the grid spacing (float)
        :return: a list of namedtuple containing a value (the LHS of the equation) and a list of values and positions to
                 insert into the matrix
        """
        raise NotImplementedError

    def constraint_symmetry(self, i_node: int, direction: int, h: float) -> List[constraint_def]:
        """
        Creates the equation that defines a symmetry constraint.

        :param i_node: the index of the node to constrain
        :param h: the grid spacing (float)
        :param direction: -1 if the left (negative) side of the structure isn't modeled; +1 otherwise
        :return: a list of namedtuple containing a value (the LHS of the equation) and a list of values and positions to
                 insert into the matrix
        """
        raise NotImplementedError

    def constraint_free_end(self, i_node: int, direction: int, h: float) -> List[constraint_def]:
        """
        Creates the equation that defines a free end. For many models, this will return an empty list.

        :param i_node: the index of the node to constrain
        :param h: the grid spacing (float)
        :param direction: -1 if the left (negative) side of the structure isn't modeled; +1 otherwise
        :return: a list of namedtuple containing a value (the LHS of the equation) and a list of values and positions to
                 insert into the matrix
        """
        raise NotImplementedError

    @staticmethod
    def result_w(z: float, terms: List[List[float]]) -> float:
        raise NotImplementedError

    @staticmethod
    def result_theta(z: float, terms: List[List[float]]) -> float:
        raise NotImplementedError

    @staticmethod
    def result_dwdx(z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of the vertical displacement with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        raise NotImplementedError

    @staticmethod
    def result_d2wdx2(z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of second derivative of the vertical displacement with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        raise NotImplementedError

    def result_dthetadx(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of theta with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        raise NotImplementedError

    def result_dtheta2dx2(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of second derivative of theta with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        raise NotImplementedError

    def result_phi(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of phi.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        raise NotImplementedError

    def result_dphidx(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of phi with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        raise NotImplementedError

    def result_u(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of the horizontal deflection

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        raise NotImplementedError

    def result_shear_strain(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the shear strain at a particular position along the beam and a particular elevation

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        raise NotImplementedError

    def result_shear_stress(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the shear stress at a particular position along the beam and a particular elevation

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        raise NotImplementedError

    def result_normal_stress(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the normal stress at a particular position along the beam and a particular elevation

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        raise NotImplemented


class PresentModel(Model):
    def __init__(self, prop: Laminate):
        super().__init__(prop, False)

        z_na = self.prop.z_na

        self._c_w1c = sum([
            p.e * (z_na * (p.zeta_top - p.zeta_bot) + 1 / 2 * (p.zeta_top ** 2 - p.zeta_bot ** 2))
            for p in self.prop.plies
        ]) / sum([
            p.e * (p.zeta_top - p.zeta_bot)
            for p in self.prop.plies
        ])

        self._c_w11 = sum([
            p.e * (-p.d_0 * (p.zeta_top - p.zeta_bot)
                   - 1 / 2 * p.d_1 * (p.zeta_top ** 2 - p.zeta_bot ** 2)
                   - 1 / 4 * p.d_3 * (p.zeta_top ** 4 - p.zeta_bot ** 4))
            for p in self.prop.plies
        ]) / sum([
            p.e * (p.zeta_top - p.zeta_bot)
            for p in self.prop.plies
        ])

        self._pc = self.prop.width * sum([
                p.e * ((z_na - self._c_w1c) ** 2 * (p.zeta_top - p.zeta_bot)
                       + (z_na - self._c_w1c) * (p.zeta_top ** 2 - p.zeta_bot ** 2)
                       + 1 / 3 * (p.zeta_top ** 3 - p.zeta_bot ** 3))
                for p in self.prop.plies
            ])

        self._p1 = self.prop.width / self.prop.calc_ei() * sum([
            p.e * (2 * (self._c_w1c - z_na) * (p.d_0 + self._c_w11) * (p.zeta_top - p.zeta_bot)
                   + (p.d_1 * (self._c_w1c - z_na) - p.d_0 - self._c_w11) * (p.zeta_top ** 2 - p.zeta_bot ** 2)
                   - 2 / 3 * p.d_1 * (p.zeta_top ** 3 - p.zeta_bot ** 3)
                   + 1 / 2 * p.d_3 * (self._c_w1c - z_na) * (p.zeta_top ** 4 - p.zeta_bot ** 4)
                   - 2 / 5 * p.d_3 * (p.zeta_top ** 5 - p.zeta_bot ** 5))
            for p in self.prop.plies
        ])

        self._p2 = self.prop.width / self.prop.calc_ei() ** 2 * sum([
            p.e * ((p.d_0 + self._c_w11) ** 2 * (p.zeta_top - p.zeta_bot)
                   + p.d_1 * (p.d_0 + self._c_w11) * (p.zeta_top ** 2 - p.zeta_bot ** 2)
                   + 1 / 3 * p.d_1 ** 2 * (p.zeta_top ** 3 - p.zeta_bot ** 3)
                   + 1 / 2 * p.d_3 * (p.d_0 + self._c_w11) * (p.zeta_top ** 4 - p.zeta_bot ** 4)
                   + 2 / 5 * p.d_1 * p.d_3 * (p.zeta_top ** 5 - p.zeta_bot ** 5)
                   + 1 / 7 * p.d_3 ** 2 * (p.zeta_top ** 7 - p.zeta_bot ** 7))
            for p in self.prop.plies
        ])

        self._nc = -self.prop.width * sum([
            p.e * (1 / 2 * (self._c_w1c - z_na) * (p.zeta_top ** 2 - p.zeta_bot ** 2)
                   - 1 / 3 * (p.zeta_top ** 3 - p.zeta_bot ** 3))
            for p in self.prop.plies
        ])

        self._n1 = -self.prop.width / self.prop.calc_ei() * sum([
            p.e * (1 / 2 * (p.d_0 + self._c_w11) * (p.zeta_top ** 2 - p.zeta_bot ** 2)
                   + 1 / 3 * p.d_1 * (p.zeta_top ** 3 - p.zeta_bot ** 3)
                   + 1 / 5 * p.d_3 * (p.zeta_top ** 5 - p.zeta_bot ** 5))
            for p in self.prop.plies
        ])

        self._validate()

    @property
    def dof(self) -> int:
        return 2

    @property
    def equation_terms(self) -> List[List[List[Union[float, None]]]]:
        return [
            [
                [None, None, -self.prop.calc_ei()],
                [None, self.prop.calc_ei() - self.calc_beta()]
            ],
            [
                [None],
                [self.calc_shear_correction() * self.prop.calc_ag(), None, -self.calc_beta()]
            ]
        ]

    def calc_beta(self) -> float:
        """
        Calculates the value of beta for the laminate

        :return: a float
        """
        kag = self.calc_shear_correction() * self.prop.calc_ag()
        return self._nc + self._n1 * kag

    def calc_shear_correction(self) -> float:
        """
        Calculates the shear correction factor based on the present model
        :return: a float
        """
        return ((np.sqrt(4 * self._p2 * (self._nc - self._pc) + (self._n1 - self._p1) ** 2) + self._n1 - self._p1) /
                (2 * self._p2)) / self.prop.calc_ag()

    def create_force_vector(self, n_model_equations: int, moment: np.array, transverse_force: np.array,
                            h: float) -> np.array:
        f = np.zeros((n_model_equations, 1))
        f[0:(len(moment) * 2):2] = moment
        f[1:(len(transverse_force) * 2 + 1):2] = transverse_force
        return f

    def constraint_transverse_force(self, i_node: int, h: float, force_magnitude: float) -> List[constraint_def]:
        return []

    def constraint_pinned(self, i_node: int, h: float) -> List[constraint_def]:
        return [constraint_def(
            value=0.,
            cv_list=list([
                (((i_node + fdc.pos) * self.dof + 0), fdc.value)
                for fdc in _finite_difference_coefficients(diff_order=0, stencil_width=0, h=h)
            ])
        )]

    def constraint_symmetry(self, i_node: int, direction: int, h: float) -> List[constraint_def]:
        return [
            # dw/dx = 0
            constraint_def(
                value=0.,
                cv_list=list([
                    ((i_node + fdc.pos) * self.dof, fdc.value)
                    for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
                ])
            ),
            # theta=0 at plane of symmetry
            constraint_def(
                value=0.,
                cv_list=list([
                    ((i_node + fdc.pos) * self.dof + 1, fdc.value)
                    for fdc in _finite_difference_coefficients(diff_order=0, stencil_width=0, h=h)
                ])
            )
        ]

    def constraint_free_end(self, i_node: int, direction: int, h: float) -> List[constraint_def]:
        # this sets d(theta)/dx = 0
        return [constraint_def(
            value=0.,
            cv_list=list([
                ((i_node + fdc.pos) * self.dof + 1, fdc.value)
                for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
            ])
        )]

    @staticmethod
    def result_w(z: float, terms: List[List[float]]) -> float:
        return terms[0][0]

    @staticmethod
    def result_theta(z: float, terms: List[List[float]]) -> float:
        return terms[1][0]

    @staticmethod
    def result_dwdx(z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of the vertical displacement with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][1]

    @staticmethod
    def result_d2wdx2(z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of second derivative of the vertical displacement with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][1]

    def result_dthetadx(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of theta with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[1][1]

    def result_dtheta2dx2(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of second derivative of theta with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[1][2]

    def result_phi(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of phi.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][1] - terms[1][0]

    def result_dphidx(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of phi with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][2] - terms[1][1]

    def result_u(self, z: float, terms: List[List[float]]) -> float:
        dwdx = terms[0][1]
        theta = terms[1][0]
        u_b = self.prop.calc_u_b(z=z, phi=dwdx - theta)
        u_w1 = self.result_u_w1(z, terms)
        return u_b + u_w1

    def _calc_r_w1(self, zeta: float) -> float:
        r_w1 = 0
        for p in self.prop.plies:
            if p.zeta_bot <= zeta <= p.zeta_bot + p.t:
                r_w1 = p.d_0 + p.d_1 * zeta + p.d_3 * zeta ** 3
        return r_w1

    def _calc_c_uw1(self) -> float:
        return self._c_w1c + self.calc_shear_correction() * self.prop.calc_ag() / self.prop.calc_ei() * self._c_w11

    def result_u_w1(self, z: float, terms: List[List[float]]) -> float:
        theta = terms[1][0]

        assert z >= 0
        assert z <= self.prop.plies[-1].z + self.prop.plies[-1].t / 2
        zeta = z - self.prop.z_na

        return theta * (
                self.calc_shear_correction() * self.prop.calc_ag() / self.prop.calc_ei() * self._calc_r_w1(zeta=zeta)
                - (zeta + self.prop.z_na) + self._calc_c_uw1()
        )

    def result_shear_strain(self, z: float, terms: List[List[float]]) -> float:
        return self.result_shear_stress(z, terms) / self.prop.get_g(z)

    _t_term = collections.namedtuple("T_Terms", ["t1", "t2", "t4"])

    def _t_terms(self, p: _Lamina, kagei: float, c_uw1: float) -> _t_term:
        t1 = p.e * (kagei * p.d_0 - self.prop.z_na + c_uw1)
        t2 = p.e / 2 * (kagei * p.d_1 - 1)
        t4 = p.e / 4 * (kagei * p.d_3)
        return self._t_term(t1, t2, t4)

    def result_shear_stress(self, z: float, terms: List[List[float]]) -> float:
        assert z >= 0
        assert z <= self.prop.plies[-1].z + self.prop.plies[-1].t / 2

        theta = terms[1][0]
        d2theta_dx2 = terms[1][2]

        tau_b = (theta * self.prop.calc_qvk(z) * self.calc_shear_correction() * self.prop.calc_ag() /
                 (self.prop.calc_ei() * self.prop.width))

        kagei = self.calc_shear_correction() * self.prop.calc_ag() / self.prop.calc_ei()
        c_uw1 = self._c_w1c + kagei * self._c_w11
        zeta = z - self.prop.z_na
        tck = 0
        for p in self.prop.plies:
            if p.zeta_top < zeta:
                t = self._t_terms(p, kagei, c_uw1)
                tck += (t.t1 * (p.zeta_top - p.zeta_bot)
                        + t.t2 * (p.zeta_top ** 2 - p.zeta_bot ** 2)
                        + t.t4 * (p.zeta_top ** 4 - p.zeta_bot ** 4))
            if p.zeta_bot <= zeta <= p.zeta_bot + p.t:
                t = self._t_terms(p, kagei, c_uw1)
                tck += -t.t1 * p.zeta_bot - t.t2 * p.zeta_bot ** 2 - t.t4 * p.zeta_bot ** 4

                return tau_b - d2theta_dx2 * (
                    tck + t.t1 * zeta + t.t2 * zeta ** 2 + t.t4 * zeta ** 4
                )

        return 0.  # Should never get here

    def result_normal_stress(self, z: float, terms: List[List[float]]) -> float:
        d2wdx2 = terms[0][2]
        dtheta_dx = terms[1][1]
        sigma_b = self.prop.calc_sigma_b(z=z, dphi_dx=d2wdx2 - dtheta_dx)

        assert z >= 0
        assert z <= self.prop.plies[-1].z + self.prop.plies[-1].t / 2
        zeta = z - self.prop.z_na

        return sigma_b + dtheta_dx * self.prop.get_e(z) * (
                self.calc_shear_correction() * self.prop.calc_ag() / self.prop.calc_ei() * self._calc_r_w1(zeta=zeta)
                - (zeta + self.prop.z_na) + self._calc_c_uw1())


class Timoshenko(Model):
    def __init__(self, prop: Laminate, fixed_shear_correction=None, do_validate: bool = True):
        self.fixed_shear_correction = fixed_shear_correction
        super().__init__(prop, do_validate)

    @property
    def dof(self) -> int:
        return 2

    @property
    def equation_terms(self) -> List[List[List[Union[float, None]]]]:
        return [
            [
                [None, None, -self.prop.calc_ei()],
                [None, self.prop.calc_ei()]
            ],
            [
                [None],
                [self.calc_shear_correction() * self.prop.calc_ag()]
            ]
        ]

    def calc_shear_correction(self) -> float:
        """
        Calculates the shear correction factor for the laminate
        :return: a float
        """
        if self.fixed_shear_correction is not None:
            return self.fixed_shear_correction
        return self.prop.calc_ei() ** 2 / self.prop.calc_ag() * self.prop.width / sum([
            (
                p.a_0 ** 2 * p.t
                + 2 / 3 * p.a_0 * p.a_2 * p.t * (p.t ** 2 + 3 * p.t * p.zeta_bot + 3 * p.zeta_bot ** 2)
                + 1 / 5 * p.a_2 ** 2 * p.t * (
                        p.t ** 4
                        + 5 * p.t ** 3 * p.zeta_bot
                        + 10 * p.t ** 2 * p.zeta_bot ** 2
                        + 10 * p.t * p.zeta_bot ** 3
                        + 5 * p.zeta_bot ** 4
                )
             ) / p.g
            for p in self.prop.plies
        ])

    def create_force_vector(self, n_model_equations: int, moment: np.array, transverse_force: np.array,
                            h: float) -> np.array:
        f = np.zeros((n_model_equations, 1))
        f[0:(len(moment) * 2):2] = moment
        f[1:(len(transverse_force) * 2 + 1):2] = transverse_force
        return f

    def constraint_transverse_force(self, i_node: int, h: float, force_magnitude: float) -> List[constraint_def]:
        return []

    def constraint_pinned(self, i_node: int, h: float) -> List[constraint_def]:
        return [constraint_def(
            value=0.,
            cv_list=list([
                (((i_node + fdc.pos) * self.dof + 0), fdc.value)
                for fdc in _finite_difference_coefficients(diff_order=0, stencil_width=0, h=h)
            ])
        )]

    def constraint_symmetry(self, i_node: int, direction: int, h: float) -> List[constraint_def]:
        return [
            # phi = 0
            # 0 = dw/dx - theta
            constraint_def(
                value=0.,
                cv_list=list([
                    ((i_node + fdc.pos) * self.dof, fdc.value)
                    for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
                ]) + list([
                    ((i_node + fdc.pos) * self.dof + 1, -fdc.value)
                    for fdc in _finite_difference_coefficients(diff_order=0, stencil_width=0, h=h)
                ])
            ),
            # set theta at virtual node equal to theta at given node
            constraint_def(
                value=0.,
                cv_list=[
                    (i_node * self.dof + 1, 1.),
                    ((i_node + direction) * self.dof + 1, -1.)
                ]
            )
        ]

    def constraint_free_end(self, i_node: int, direction: int, h: float) -> List[constraint_def]:
        # set theta at virtual node equal to theta at given node
        return [constraint_def(
            value=0.,
            cv_list=[
                (i_node * self.dof + 1, 1.),
                ((i_node + direction) * self.dof + 1, -1.)
            ]
        )]

    @staticmethod
    def result_w(z: float, terms: List[List[float]]) -> float:
        return terms[0][0]

    @staticmethod
    def result_theta(z: float, terms: List[List[float]]) -> float:
        return terms[1][0]

    @staticmethod
    def result_dwdx(z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of the vertical displacement with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][1]

    @staticmethod
    def result_d2wdx2(z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of second derivative of the vertical displacement with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][1]

    def result_dthetadx(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of theta with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[1][1]

    def result_dtheta2dx2(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of second derivative of theta with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[1][2]

    def result_phi(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of phi.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][1] - terms[1][0]

    def result_dphidx(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of phi with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][2] - terms[1][1]

    def result_u(self, z: float, terms: List[List[float]]) -> float:
        dwdx = terms[0][1]
        theta = terms[1][0]
        u_b = self.prop.calc_u_b(z=z, phi=dwdx - theta)
        u_w1 = 0
        return u_b + u_w1

    def result_shear_strain(self, z: float, terms: List[List[float]]) -> float:
        theta = terms[1][0]

        return (theta * self.prop.calc_qvk(z) * self.calc_shear_correction() * self.prop.calc_ag() /
                (self.prop.calc_ei() * self.prop.width * self.prop.get_g(z)))

    def result_shear_stress(self, z: float, terms: List[List[float]]) -> float:
        return self.result_shear_strain(z, terms) * self.prop.get_g(z)

    def result_normal_stress(self, z: float, terms: List[List[float]]) -> float:
        d2wdx2 = terms[0][2]
        dtheta_dx = terms[1][1]
        return self.prop.calc_sigma_b(z=z, dphi_dx=d2wdx2 - dtheta_dx)


class EulerBernoilli(Model):
    @property
    def dof(self) -> int:
        return 1

    @property
    def equation_terms(self) -> List[List[List[Union[float, None]]]]:
        return [
            [
                [None, None, -self.prop.calc_ei()]
            ]
        ]

    def create_force_vector(self, n_model_equations: int, moment: np.array, transverse_force: np.array,
                            h: float) -> np.array:
        f = np.zeros((n_model_equations, 1))
        f[0:len(moment):1] = moment
        return f

    def constraint_transverse_force(self, i_node: int, h: float, force_magnitude: float) -> List[constraint_def]:
        return []

    def constraint_pinned(self, i_node: int, h: float) -> List[constraint_def]:
        return [constraint_def(
            value=0.,
            cv_list=list([
                (((i_node + fdc.pos) * self.dof + 0), fdc.value)
                for fdc in _finite_difference_coefficients(diff_order=0, stencil_width=0, h=h)
            ])
        )]

    def constraint_symmetry(self, i_node: int, direction: int, h: float) -> List[constraint_def]:
        return [constraint_def(
            value=0.,
            cv_list=list([
                ((i_node + fdc.pos) * self.dof, fdc.value)
                for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
            ])
        )]

    def constraint_free_end(self, i_node: int, direction: int, h: float) -> List[constraint_def]:
        return []

    @staticmethod
    def result_w(z: float, terms: List[List[float]]) -> float:
        return terms[0][0]

    @staticmethod
    def result_theta(z: float, terms: List[List[float]]) -> float:
        return 0

    @staticmethod
    def result_dwdx(z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of the vertical displacement with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][1]

    @staticmethod
    def result_d2wdx2(z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of second derivative of the vertical displacement with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][1]

    def result_dthetadx(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of theta with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return 0

    def result_dtheta2dx2(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of second derivative of theta with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return 0

    def result_phi(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of phi.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][1]

    def result_dphidx(self, z: float, terms: List[List[float]]) -> float:
        """
        Returns the value of first derivative of phi with respect to position.

        See Beam.get_computed_result for additional details.

        :param z: the elevation of interest. Included only to match the generic signature
        :param terms: a list of lists of terms at the location of interest
        :return: the computed value
        """
        return terms[0][2]

    def result_u(self, z: float, terms: List[List[float]]) -> float:
        return self.prop.calc_u_b(z=z, phi=terms[0][1])

    def result_shear_strain(self, z: float, terms: List[List[float]]) -> float:
        return 0

    def result_shear_stress(self, z: float, terms: List[List[float]]) -> float:
        return 0

    def result_normal_stress(self, z: float, terms: List[List[float]]) -> float:
        return self.prop.calc_sigma_b(z=z, dphi_dx=terms[0][2])


@dataclass(init=False)
class _ZigZagCoefficients:
    a11: float
    b12: float
    d11: float
    b13: float
    d12: float
    d22: float
    q11: float
    q12: float
    q22: float


class ZigZag(Model):
    """
    This class based on NASA/TP-2007-215086
    """

    def __init__(self, prop: Laminate, do_validate: bool = False):
        super().__init__(prop, do_validate=False)
        self.u_k = []
        self.beta = []
        self.g_laminate = 0.
        self.stiffness_coefficients = self._compute_stiffness_coefficients()
        if do_validate:
            self._validate()

    @property
    def dof(self) -> int:
        return 4  # Since B13 is non-zero for non-symmetric laminates, we need to keep u as a DOF

    def _compute_stiffness_coefficients(self) -> _ZigZagCoefficients:
        sc = _ZigZagCoefficients()
        w = self.prop.width
        lam_0 = 1  # for the "refined" zig-zag model

        sc.a11 = w * sum([p.e * (p.z_ref_mid_top - p.z_ref_mid_bot) for p in self.prop.plies])
        sc.b12 = w * sum([p.e * (p.z_ref_mid_top ** 2 - p.z_ref_mid_bot ** 2) / 2 for p in self.prop.plies])
        sc.d11 = w * sum([p.e * (p.z_ref_mid_top ** 3 - p.z_ref_mid_bot ** 3) / 3 for p in self.prop.plies])

        t = self.prop.thickness
        self.g_laminate = (t / 2) / sum([(p.t / 2) / p.g for p in self.prop.plies])
        # This model defines u(k), which are the incremental displacements at each layer interface. u(0)=u(N)=0.
        self.u_k = [0] * (len(self.prop.plies) + 1)
        for k, p in zip(range(1, len(self.prop.plies) + 1, 1), self.prop.plies):
            self.u_k[k] = self.u_k[k - 1] + p.t * (self.g_laminate / p.g - 1)

        # the function phi(k) linearly interpolates from u(k-1) to u(k) through the thickness of each ply

        self.beta = list([(u2 - u1) / p.t for u1, u2, p in zip(self.u_k, self.u_k[1:], self.prop.plies)])
        sc.b13 = w * sum([
            p.e * (u1 + u2) / 2 * p.t
            for u1, u2, p in zip(self.u_k, self.u_k[1:], self.prop.plies)
        ])
        sc.d12 = w * sum([
            p.e * (1 / 2 * (u - b * p.z_ref_mid_bot) * (p.z_ref_mid_top ** 2 - p.z_ref_mid_bot ** 2)
                   + 1 / 3 * b * (p.z_ref_mid_top ** 3 - p.z_ref_mid_bot ** 3))
            for u, b, p in zip(self.u_k, self.beta, self.prop.plies)
        ])
        sc.d22 = w * sum([
            p.e * (
                    (u ** 2 - 2 * b * u * p.z_ref_mid_bot + b ** 2 * p.z_ref_mid_bot ** 2) * (
                        p.z_ref_mid_top - p.z_ref_mid_bot)
                    + (u * b - b ** 2 * p.z_ref_mid_bot) * (p.z_ref_mid_top ** 2 - p.z_ref_mid_bot ** 2)
                    + b ** 2 / 3 * (p.z_ref_mid_top ** 3 - p.z_ref_mid_bot ** 3))
            for u, b, p in zip(self.u_k, self.beta, self.prop.plies)
        ])
        q = w * sum([
            p.t * p.g * (1 + b) ** 2
            for b, p in zip(self.beta, self.prop.plies)
        ])
        lam = lam_0 * (w * sum([p.g * p.t for p in self.prop.plies]) - q)
        sc.q11 = q + lam
        sc.q12 = -lam
        sc.q22 = lam
        return sc

    @property
    def equation_terms(self) -> List[List[List[Union[float, None]]]]:
        sc = self.stiffness_coefficients
        # The equations are:
        #   N,x = 0
        #   Mx,x - Vx = 0
        #   Vx,x = -q
        #   M_phi,x - V_phi = 0
        # The DOF are:
        #   u
        #   w
        #   theta
        #   psi
        return [
            [  # N,x = 0
                [None, None, sc.a11],  # u
                [None],  # w
                [None, None, sc.b12],  # theta
                [None, None, sc.b13]  # psi
            ],
            [  # Mx,x - Vx = 0
                [None, None, sc.b12],  # u
                [None, -sc.q11],  # w
                [-sc.q11, None, sc.d11],  # theta
                [-sc.q12, None, sc.d12]  # psi
            ],
            [  # Vx,x = q
                [None],  # u
                [None, None, sc.q11],  # w
                [None, sc.q11],  # theta
                [None, sc.q12]  # psi
            ],
            [  # M_phi,x - V_phi
                [None, None, sc.b13],  # u
                [None, -sc.q12],  # w
                [-sc.q12, None, sc.d12],  # theta
                [-sc.q22, None, sc.d22]  # psi
            ]
        ]

    def create_force_vector(self, n_model_equations: int, moment: np.array, transverse_force: np.array,
                            h: float) -> np.array:
        f = np.zeros((n_model_equations, 1))
        # The first equation is all zeros (N,x = 0)
        # The second equation is all zeros (Mx,x - Vx = 0)
        # The third equation is all zeros if the distributed load is zero (Vx,x = -q)
        q = np.zeros_like(transverse_force)
        for i, (v1, v2) in enumerate(zip(transverse_force, transverse_force[1::])):
            q[i + 1] = (v2 - v1) / h
        f[2:(len(transverse_force) * 4 + 2):4] = q
        # The last equation is all zeros (M_phi,x - V_phi = 0)
        return f

    def constraint_transverse_force(self, i_node: int, h: float, force_magnitude: float) -> List[constraint_def]:
        sc = self.stiffness_coefficients
        cv_list = []
        cv_list.extend([  # q_11 * theta
            (((i_node + fdc.pos) * self.dof + 2), fdc.value * sc.q11)
            for fdc in _finite_difference_coefficients(diff_order=0, stencil_width=0, h=h)
        ])
        cv_list.extend([  # q_11 * w,x
            (((i_node + fdc.pos) * self.dof + 1), fdc.value * sc.q11)
            for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
        ])
        cv_list.extend([  # q_12 * psi
            (((i_node + fdc.pos) * self.dof + 3), fdc.value * sc.q12)
            for fdc in _finite_difference_coefficients(diff_order=0, stencil_width=0, h=h)
        ])
        return [
            # Vx = v
            constraint_def(
                value=force_magnitude,
                cv_list=cv_list
            )
        ]

    def constraint_pinned(self, i_node: int, h: float) -> List[constraint_def]:
        return [
            # set w = 0 at pin
            constraint_def(
                value=0.,
                cv_list=list([
                    (((i_node + fdc.pos) * self.dof + 1), fdc.value)
                    for fdc in _finite_difference_coefficients(diff_order=0, stencil_width=0, h=h)
                ])
            )
        ]

    def constraint_symmetry(self, i_node: int, direction: int, h: float) -> List[constraint_def]:
        return [
            # set u = 0
            constraint_def(
                value=0.,
                cv_list=list([
                    (((i_node + fdc.pos) * self.dof + 0), fdc.value)
                    for fdc in _finite_difference_coefficients(diff_order=0, stencil_width=0, h=h)
                ])
            ),
            # theta=0 at plane of symmetry
            constraint_def(
                value=0.,
                cv_list=list([
                    ((i_node + fdc.pos) * self.dof + 2, fdc.value)
                    for fdc in _finite_difference_coefficients(diff_order=0, stencil_width=0, h=h)
                ])
            ),
            # psi = 0 at plane of symmetry
            constraint_def(
                value=0,
                cv_list=list([
                    ((i_node + fdc.pos) * self.dof + 3, fdc.value)
                    for fdc in _finite_difference_coefficients(diff_order=0, stencil_width=0, h=h)
                ])
            )
        ]

    def constraint_free_end(self, i_node: int, direction: int, h: float) -> List[constraint_def]:
        sc = self.stiffness_coefficients

        # Need to set M_phi = 0
        cv_m_phi = []
        cv_m_phi.extend([  # b_13 * u,x
            (((i_node + fdc.pos) * self.dof + 0), fdc.value * sc.b13)
            for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
        ])
        cv_m_phi.extend([  # d_12 * theta,x
            (((i_node + fdc.pos) * self.dof + 2), fdc.value * sc.d12)
            for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
        ])
        cv_m_phi.extend([  # d_22 * phi,x
            (((i_node + fdc.pos) * self.dof + 3), fdc.value * sc.d22)
            for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
        ])

        # Need to set M_x = 0
        cv_m_x = []
        cv_m_x.extend([  # b_12 * u,x
            (((i_node + fdc.pos) * self.dof + 0), fdc.value * sc.b12)
            for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
        ])
        cv_m_x.extend([  # d_11 * theta,x
            (((i_node + fdc.pos) * self.dof + 2), fdc.value * sc.d11)
            for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
        ])
        cv_m_x.extend([  # d_12 * phi,x
            (((i_node + fdc.pos) * self.dof + 3), fdc.value * sc.d12)
            for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
        ])

        # Need to set N = 0
        cv_n = []
        cv_n.extend([  # a11 * u,x
            (((i_node + fdc.pos) * self.dof + 0), fdc.value * sc.a11)
            for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
        ])
        cv_n.extend([  # b12 * theta,x
            (((i_node + fdc.pos) * self.dof + 2), fdc.value * sc.b12)
            for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
        ])
        cv_n.extend([  # b13 * psi,x
            (((i_node + fdc.pos) * self.dof + 3), fdc.value * sc.b13)
            for fdc in _finite_difference_coefficients(diff_order=1, stencil_width=1, h=h)
        ])

        return [
            # set M_phi = 0
            constraint_def(
                value=0.,
                cv_list=cv_m_phi
            ),
            # set M_x = 0
            constraint_def(
                value=0.,
                cv_list=cv_m_x
            ),
            # set N = 0
            constraint_def(
                value=0.,
                cv_list=cv_n
            )
        ]

    @staticmethod
    def result_w(z: float, terms: List[List[float]]) -> float:
        return terms[1][0]

    @staticmethod
    def result_theta(z: float, terms: List[List[float]]) -> float:
        return terms[2][0]

    @staticmethod
    def result_dwdx(z: float, terms: List[List[float]]) -> float:
        return terms[1][1]

    @staticmethod
    def result_d2wdx2(z: float, terms: List[List[float]]) -> float:
        return terms[1][2]

    def result_dthetadx(self, z: float, terms: List[List[float]]) -> float:
        pass

    def result_dtheta2dx2(self, z: float, terms: List[List[float]]) -> float:
        pass

    def result_phi(self, z: float, terms: List[List[float]]) -> float:
        pass

    def result_dphidx(self, z: float, terms: List[List[float]]) -> float:
        pass

    def result_psi(self, z: float, terms: List[List[float]]) -> float:
        return terms[3][0]

    def result_n_x(self, z: float, terms: List[List[float]]) -> float:
        sc = self.stiffness_coefficients
        return sc.a11 * terms[0][1] + sc.b12 * terms[2][1] + sc.b13 * terms[3][1]

    def result_m_x(self, z: float, terms: List[List[float]]) -> float:
        sc = self.stiffness_coefficients
        return sc.b12 * terms[0][1] + sc.d11 * terms[2][1] + sc.d12 * terms[3][1]

    def result_m_phi(self, z: float, terms: List[List[float]]) -> float:
        sc = self.stiffness_coefficients
        return sc.b13 * terms[0][1] + sc.d12 * terms[2][1] + sc.d22 * terms[3][1]

    def result_v_x(self, z: float, terms: List[List[float]]) -> float:
        sc = self.stiffness_coefficients
        return sc.q11 * (terms[2][0] + terms[1][1]) + sc.q12 * terms[3][0]

    def result_v_phi(self, z: float, terms: List[List[float]]) -> float:
        sc = self.stiffness_coefficients
        return sc.q12 * (terms[2][0] + terms[1][1]) + sc.q22 * terms[3][0]

    def result_u(self, z: float, terms: List[List[float]]) -> float:
        z = z - self.prop.thickness / 2
        ux = terms[0][0]
        theta = terms[2][0]
        psi = terms[3][0]

        phi = 0
        for p, u1, u2 in zip(self.prop.plies, self.u_k, self.u_k[1:]):
            if p.z_ref_mid_bot <= z <= p.z_ref_mid_top:
                phi = u1 + (u2 - u1) * (z - p.z_ref_mid_bot) / p.t

        return ux + z * theta + phi * psi

    def result_shear_strain(self, z: float, terms: List[List[float]]) -> float:
        gamma = terms[1][1] + terms[2][0]  # w,x + theta
        psi = terms[3][0]
        eta = gamma - terms[3][0]
        z = z - self.prop.thickness / 2
        for p in self.prop.plies:
            if p.z_ref_mid_bot <= z <= p.z_ref_mid_top:
                return eta + self.g_laminate / p.g * psi
        return 0

    def result_shear_stress(self, z: float, terms: List[List[float]]) -> float:
        return self.prop.get_g(z) * self.result_shear_strain(z, terms)

    def result_normal_stress(self, z: float, terms: List[List[float]]) -> float:
        z = z - self.prop.thickness / 2

        for p, u1, u2 in zip(self.prop.plies, self.u_k, self.u_k[1:]):
            if p.z_ref_mid_bot <= z <= p.z_ref_mid_top:
                u = u1 + (u2 - u1) * (z - p.z_ref_mid_bot) / p.t
                strain = terms[0][1] + z * terms[2][1] + u * terms[3][1]
                return strain * p.e
        return 0  # should never reach here


def _partition_plies(lam: Laminate, num_partitions: int, s: float) -> Laminate:
    partitioned_laminate = copy.deepcopy(lam)
    new_plies = []
    for p in lam.plies:
        for i in range(num_partitions):
            p3 = copy.deepcopy(p)
            p3.t = p.t / num_partitions
            new_plies.append(p3)
    partitioned_laminate.plies = new_plies
    for p in partitioned_laminate.plies:
        epsilon = s / (partitioned_laminate.thickness / 2) * (
            p.z_ref_mid_bot ** 2 + p.z_ref_mid_top ** 2 + p.z_ref_mid_bot * p.z_ref_mid_top
        )
        p.g *= 1 + epsilon
    partitioned_laminate.plies = partitioned_laminate.plies  # force an update
    return partitioned_laminate


class ZigZagH(ZigZag):
    """
    This class based on NASA/TP-2007-215086 and NASA/TP-2010-216214
    """

    def __init__(self, prop: Laminate, ply_partitions: int,
                 s: float = 0, do_validate: bool = False):
        prop = _partition_plies(prop, ply_partitions, s)
        super().__init__(prop, do_validate=do_validate)


class Beam:
    def __init__(self, n_dof: int):
        self.model = None
        self._n_dof = n_dof
        self.h = 0.
        self.moment = np.zeros((1, 1))
        self.shear = np.zeros((1, 1))
        self.x_node = np.zeros(1)
        self._delta = np.zeros((1, 1))
        self._col_offset = 0

    @property
    def constraints(self) -> List[constraint_def]:
        """
        A property which returns a list of all the constraints applied to the beam.
        :return: a list of `constraint_def`s
        """
        return NotImplemented

    def _generate_terms_for_node(self, i_node: int,
                                 terms: List[List[Union[float, None]]]) -> Tuple[List[int], List[float]]:
        c = []
        val = []

        for cur_dof, cur_term in enumerate(terms):
            for derivative_order, term_value in enumerate(cur_term):
                if term_value is not None:
                    for fdc in _finite_difference_coefficients(derivative_order,
                                                               stencil_width=min(derivative_order, 1), h=self.h):
                        c.append((i_node + fdc.pos) * self._n_dof + cur_dof)
                        val.append(fdc.value * term_value)
        return c, val

    def solve(self, *, print_diagnostics: bool = False) -> None:
        """
        Solves the current beam model.

        :param print_diagnostics: a boolean indicating whether to print diagnostics for the matrix
        :return: None
        """
        equation_terms = self.model.equation_terms

        r = []
        c = []
        val = []

        for i_node, x in enumerate(self.x_node):
            for i_eq, eq in enumerate(equation_terms):
                cur_c, cur_v = self._generate_terms_for_node(i_node, eq)
                r.extend([i_node * self._n_dof + i_eq] * len(cur_v))
                c.extend(cur_c)
                val.extend(cur_v)

        self._col_offset = -min(c)
        c = list([ci + self._col_offset for ci in c])
        n_cols = max(c) + 1

        f = self.model.create_force_vector(
            n_model_equations=n_cols,
            moment=self.moment,
            transverse_force=self.shear,
            h=self.h
        )

        # Add in constraints as new equations
        constraint_row_offset = max(r) + 1
        for i_c, const in enumerate(self.constraints):
            for cv in const.cv_list:
                r.append(i_c + constraint_row_offset)
                c.append(cv[0] + self._col_offset)
                val.append(cv[1])
            f[i_c + constraint_row_offset, 0] = const.value

        assert max(r) == max(c)
        assert min(r) == 0
        assert min(c) == 0

        k = scipy.sparse.coo_matrix((val, (r, c)), shape=(n_cols, n_cols))
        k = k.tocsr()

        if print_diagnostics:
            matrix_diagnostics(k)

        self._delta = scipy.sparse.linalg.spsolve(k, f)

    @property
    def w(self):
        w = np.array([self.get_computed_result("w", i_node, z=0) for i_node, _ in enumerate(self.x_node)])
        return w

    @property
    def theta(self):
        theta = np.array([self.get_computed_result("theta", i_node, z=0) for i_node, _ in enumerate(self.x_node)])
        return theta

    def plot_shear_moment_diagram(self):
        """
        Plots the shear and moment applied to the beam versus the position along the length of the beam
        :return: None
        """
        fig = plt.figure()
        ax = fig.add_subplot(2, 1, 1)
        ax.plot(self.x_node, self.shear)
        ax.set_ylabel("Shear")
        ax = fig.add_subplot(2, 1, 2)
        ax.plot(self.x_node, self.moment)
        ax.set_ylabel("Moment")
        return ax

    def plot_deformed_shape(self, ax=None, **kwargs):
        """
        Plots the deformed shape of the beam
        :return: None
        """
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
        ax.plot(self.x_node, self.w, **kwargs)
        return ax

    def plot_theta(self, ax=None, **kwargs):
        """
        Plots the value of theta along the length of the beam
        :return: None
        """
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
        ax.plot(self.x_node, self.theta, **kwargs)
        return ax

    def plot_result_along_beam(self, result_name: str, z: float, *, scale_result: float = 1, scale_x: float = 1,
                               ax=None, **kwargs):
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
        x = self.x_node * scale_x
        y = [
            self.get_computed_result(result_name, i, z) * scale_result
            for i, _ in enumerate(self.x_node)
        ]
        ax.plot(x, y, **kwargs)
        return ax

    def plot_result_through_thickness(self, result_name: str, i_node: int, *,
                                      scale_result: float = 1, scale_z: float = 1, ax=None, **kwargs):
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
        z = np.linspace(0, self.model.prop.plies[-1].z + self.model.prop.plies[-1].t / 2, num=500, endpoint=False)
        res = self.get_computed_result(result_name, i_node, z) * scale_result
        ax.plot(res, z * scale_z, **kwargs)
        return ax

    def get_computed_result(self, result_name: str, i_node: int, z: Union[np.array, float]) -> np.array:
        """
        Gets a result computed by the model. This method looks for a method within the model named
        result_{result_name}. For example, specifying result_name="dwdx" will execute the method
        of the model class called result_dwdx.

        Methods must have the signature:

        result_name(self, z: float, terms: List[List[float]]) -> float:

        The terms argument is a list. Each element of the outer list corresponds with the degrees of
        freedom (w and theta). Each element of that list give the 0-th through 2-nd derivatives of the
        variable. Hence, terms[0][1] would be the first derivative of w and terms[1][2] would be the
        second derivative of theta. Similarly terms[0][0] is just w.

        If the model has a single DOF, then the outside list of terms has only a single element,
        corresponding with w.

        The get_computed_result method takes two additional arguments, the node at which the result
        is sought, and a vector of elevations at which the result should be computed.

        :param result_name: the name of the result desired
        :param i_node: the node at which the result should be evaluated
        :param z: a vector of elevations at which the result should be evaluated
        :return: a numpy.array or a float (to match z)
        """
        method_name = "result_{}".format(result_name)
        method = getattr(self.model, method_name)

        terms = list([
            list([
                sum([
                    self._delta[(i_node + fdc.pos) * self._n_dof + i_dof + self._col_offset] * fdc.value
                    for fdc in _finite_difference_coefficients(i_diff, stencil_width=min(i_diff, 1), h=self.h)
                ])
                for i_diff in range(3)
            ])
            for i_dof in range(self._n_dof)
        ])

        if type(z) is float or type(z) is int:
            return method(z=z, terms=terms)
        else:
            return np.array([
                method(z=zi, terms=terms)
                for zi in z
            ])


class ThreePointSymmetricBeam(Beam):
    """
    A sub-class of the Beam class, defining a beam supported at two locations and loaded in the middle.
    This uses a symmetry condition to simplify the model, so the actual beam solved is a beam with a symmetry
    condition applied to its left edge (and loaded with half the total load at this edge), and supported by
    a pin at some location along the beam. The beam overhangs the support.
    """
    def __init__(self, length: float, support_x: float, n_elm: int, load: float, model: Model):
        super().__init__(n_dof=model.dof)
        self.model = model
        self.length = length
        self.support_x = support_x
        self.load = load
        self.h = length / n_elm

        self.x_node = np.linspace(0, length, num=n_elm + 1)

        self.shear = np.array([[load if x <= support_x else 0 for x in self.x_node]]).T
        self.moment = np.array([[load * (x - support_x) if x <= support_x else 0 for x in self.x_node]]).T

        self._constraints = self._calc_constraints()

    def _calc_constraints(self) -> List[constraint_def]:
        i_sup = int(np.ceil(self.support_x / self.h))

        constraints = []
        # w = 0 at support
        constraints.extend(self.model.constraint_pinned(i_sup, self.h))
        # dw/dx=0 at plane of symmetry and some constraint for theta, depending on model
        constraints.extend(self.model.constraint_symmetry(0, direction=-1,  h=self.h))
        # Constraint for free end, which is needed for some models
        constraints.extend(self.model.constraint_free_end(len(self.x_node) - 1, direction=1, h=self.h))
        # The applied load
        constraints.extend(self.model.constraint_transverse_force(0, h=self.h, force_magnitude=self.load))

        return constraints

    @property
    def constraints(self) -> List[constraint_def]:
        return self._constraints


class CantileveredBeam(Beam):
    """
    """
    def __init__(self, length: float, n_elm: int, load: float, model: Model):
        super().__init__(n_dof=model.dof)
        self.model = model
        self.length = length
        self.load = load
        self.h = length / n_elm

        self.x_node = np.linspace(0, length, num=n_elm + 1)

        self.shear = np.array([[load for _ in self.x_node]]).T
        self.moment = np.array([[load * (x - length) for x in self.x_node]]).T

        self._constraints = self._calc_constraints()

    def _calc_constraints(self) -> List[constraint_def]:
        constraints = []
        # at the support
        constraints.extend(self.model.constraint_pinned(0, self.h))
        constraints.extend(self.model.constraint_symmetry(0, direction=-1, h=self.h))
        # Constraint for free end, which is needed for some models
        constraints.extend(self.model.constraint_free_end(len(self.x_node) - 1, direction=1, h=self.h))
        # The applied load
        constraints.extend(self.model.constraint_transverse_force(len(self.x_node) - 1, h=self.h,
                                                                  force_magnitude=self.load))

        return constraints

    @property
    def constraints(self) -> List[constraint_def]:
        return self._constraints
