import scipy.optimize
from typing import List
from .ply import Ply
from .ply import _Lamina


class Laminate:
    """
    This class defines the cross-section of the beam. It assumes that the beam is made from one or more
    homogeneous layer(s).
    """
    def __init__(self):
        self._plies = []
        self.width = 0.
        self.z_na = 0.
        self.zeta_bot = 0.
        self.thickness = 0.

    @property
    def plies(self):
        return self._plies

    @plies.setter
    def plies(self, value: List[Ply]):
        self._plies = list([_Lamina(p) for p in value])
        self._update_na_height()
        self._set_ply_a_constants()
        self._set_ply_d_constants()

    def _update_na_height(self) -> None:
        """
        Calculate the height of the neutral axis and update the plies with their heights relative to the bottom
        of the laminate, and also relative to the neutral axis
        :return: None
        """
        z_center = 0
        for p in self._plies:
            z_center += p.t / 2
            p.z = z_center
            z_center += p.t / 2

        self.thickness = z_center

        self.z_na = sum([
            p.e * ((p.z + p.t / 2) ** 2 - (p.z - p.t / 2) ** 2)
            for p in self._plies
        ]) / 2 / sum([
            p.e * p.t
            for p in self._plies
        ])

        for p in self._plies:
            p.zeta_center = p.z - self.z_na
            p.zeta_bot = p.zeta_center - p.t / 2
            p.zeta_top = p.zeta_center + p.t / 2
            p.z_ref_mid_bot = p.z - self.thickness / 2 - p.t / 2
            p.z_ref_mid_top = p.z - self.thickness / 2 + p.t / 2

        self.zeta_bot = self._plies[0].zeta_bot

    def _set_ply_a_constants(self) -> None:
        """
        Sets the A_0k and A_2k constants for the plies. These are used in the PresentModel

        This function must be called after _update_na_height
        :return: None
        """
        sum_term = 0
        for p in self._plies:
            p.a_0 = self.width / 2 * (p.e * p.zeta_bot ** 2 - sum_term)
            p.a_2 = -self.width / 2 * p.e
            sum_term += p.e * (p.zeta_top ** 2 - p.zeta_bot ** 2)

    def _set_ply_d_constants(self) -> None:
        """
        Sets the D_{0k}, D_{1k} and D_{3k} constants for each ply. These are used in the PresentModel

        This function must be called after _set_ply_a_constants
        :return:
        """
        sum_term = 0
        for p in self._plies:
            p.d_0 = sum_term - (p.a_0 * p.zeta_bot + p.a_2 * p.zeta_bot ** 3 / 3) / (self.width * p.g)
            p.d_1 = p.a_0 / (self.width * p.g)
            p.d_3 = p.a_2 / (3 * self.width * p.g)
            sum_term += (
                p.a_0 * (p.zeta_top - p.zeta_bot)
                + p.a_2 / 3 * (p.zeta_top ** 3 - p.zeta_bot ** 3)
                ) / (
                    self.width * p.g
            )

    def calc_ei(self) -> float:
        """
        Calculates value of the elastic modulus times the moment of inertia for the laminate
        :return: a float
        """
        return self.width * sum([p.e * ((p.zeta_center + p.t / 2) ** 3 - (p.zeta_center - p.t / 2) ** 3) / 3 for p in self._plies])

    def calc_ag(self) -> float:
        """
        Calculates the value of the shear modulus times the area for the laminate
        :return: a float
        """
        return sum([p.g * self.width * p.t for p in self._plies])

    def calc_qvk(self, z: float) -> float:
        """
        Calculates the value of Q_{kk} given a value of z, which is the height within the beam, relative to the bottom
        surface
        :param z: The height relative to the bottom surface of the beam (float)
        :return: a float
        """
        assert z >= 0
        assert z <= self._plies[-1].z + self._plies[-1].t / 2
        zeta = z - self.z_na
        for p in self._plies:
            if p.zeta_bot <= zeta <= p.zeta_bot + p.t:
                return p.a_0 + p.a_2 * zeta ** 2
        return 0

    def calc_u_b(self, z: float, phi: float) -> float:
        """
        Calculate the longitudinal deformation caused by bending, given a value of phi, at an
        elevation, z
        :param z: float
        :param phi: float
        :return: float
        """
        return -(z - self.z_na) * phi

    def calc_sigma_b(self, z: float, dphi_dx: float) -> float:
        """
        Calculate the longitudinal normal stress caused by bending, given a value of phi, at an
        elevation, z
        :param z: float
        :param phi: float
        :return: float
        """
        return -(z - self.z_na) * self.get_e(z) * dphi_dx

    def get_g(self, z: float) -> float:
        """
        Gets the value of the shear modulus at an elevation z
        :param z: z
        :return: float
        """
        for p in self._plies:
            if p.z - p.t / 2 <= z <= p.z + p.t / 2:
                return p.g
        return 0  # should never get here

    def get_e(self, z: float) -> float:
        """
        Gets the value of the elastic modulus at an elevation z
        :param z: z
        :return: float
        """
        for p in self._plies:
            if p.z - p.t / 2 <= z <= p.z + p.t / 2:
                return p.e
        return 0  # should never get here
