from unittest import TestCase
import numpy as np
import Beam


class TestBeam(TestCase):
    def test__finite_difference_coefficients(self):
        h = 10. / 100

        # Check coefficients 0-th derivative
        res = Beam.beam._finite_difference_coefficients(0, 1, h)
        for r in res:
            self.assertAlmostEqual(r.value, 1.0 if r.pos == 0 else 0)

        # Check coefficients 1st derivative. Set stencil width to 1
        res = Beam.beam._finite_difference_coefficients(1, stencil_width=1, h=h)
        expected = {
            -1: -1 / (2 * h),
            0: 0,
            1: 1 / (2 * h)
        }
        for r in res:
            self.assertAlmostEqual(r.value, expected[r.pos] if r.pos in expected else 0)

        # Check coefficients for 2nd derivative. Set stencil width to 1
        res = Beam.beam._finite_difference_coefficients(2, stencil_width=1, h=h)
        expected = {
            -1: 1 / h ** 2,
            0: -2 / h ** 2,
            1: 1 / h ** 2
        }
        for r in res:
            self.assertAlmostEqual(r.value, expected[r.pos] if r.pos in expected else 0)

    def build_lam4(self) -> Beam.Laminate:
        c0 = Beam.Ply()
        c0.e = 119000
        c0.g = 4230
        c0.t = 1

        ss = Beam.Ply()
        ss.e = 9270
        ss.g = 323
        ss.t = 1

        lam4 = Beam.Laminate()
        lam4.width = 10
        lam4.plies = [c0, ss, c0]

        return lam4

    def test_timoshenko(self):
        lam4 = self.build_lam4()
        model = Beam.Timoshenko(lam4)
        beam = Beam.ThreePointSymmetricBeam(18., 12., 1200, 50., model)
        beam.solve()

        def analytical(x: float) -> float:
            p = beam.load * 2
            a = 12
            ei = lam4.calc_ei()
            kag = model.calc_shear_correction() * lam4.calc_ag()
            v = p / 2
            m = -p * a / 2
            c1 = v / kag - v / kag
            c2 = v / (6 * ei) * a ** 3 + m / (2 * ei) * a ** 2 - 1 / kag * v * a - c1 * a
            w = 1 / kag * (v * x - v * (x - a) * np.heaviside(x - a, 1)) + \
                v / (6 * ei) * ((a - x) ** 3 * (np.heaviside(a - x, 1) - 1) - x ** 3) - \
                m / (2 * ei) * x ** 2 + \
                c1 * x + \
                c2
            return w

        w_analytical = np.array([
            analytical(xi) for xi in beam.x_node
        ])

        w_model = np.array([
            beam.get_computed_result("w", i_node, 0)
            for i_node in range(len(beam.x_node))
        ])

        for w_m, w_a in zip(w_model, w_analytical):
            self.assertAlmostEqual(w_m, w_a, delta=5e-5)

    def test_timoshenko_shear_correction(self):
        aluminum = Beam.Ply()
        aluminum.e = 73e9
        aluminum.g = 28e9
        aluminum.t = 10

        lam = Beam.Laminate()
        lam.width = 20

        lam.plies = [aluminum]
        self.assertAlmostEqual(Beam.Timoshenko(lam).calc_shear_correction(), 5 / 6)

        lam.plies = [aluminum, aluminum]
        self.assertAlmostEqual(Beam.Timoshenko(lam).calc_shear_correction(), 5 / 6)

    def test_euler_bernoulli(self):
        lam4 = self.build_lam4()
        beam = Beam.ThreePointSymmetricBeam(18., 12., 600, 50., Beam.EulerBernoilli(lam4))
        beam.solve()

        def euler_bernoulli(x: float) -> float:
            p = beam.load * 2
            a = 12
            ei = lam4.calc_ei()
            w = -p * a ** 3 / (6 * ei) + \
                p * a / (4 * ei) * x ** 2 - \
                p / (12 * ei) * x ** 3 + \
                p / (12 * ei) * (x - a) ** 3 * np.heaviside(x - a, 1)
            return w

        w_analytical = np.array([
            euler_bernoulli(xi) for xi in beam.x_node
        ])

        w_model = np.array([
            beam.get_computed_result("w", i_node, 0)
            for i_node in range(len(beam.x_node))
        ])

        for w_m, w_a in zip(w_model, w_analytical):
            self.assertAlmostEqual(w_m, w_a, delta=5e-5)
